from docker import from_env
from docker.types.services import EndpointSpec, ServiceMode

from . import Tool


class OSPRayStudio(Tool):
	def __init__(self, config, data_sources):
		super().__init__(config, data_sources)

		self.name = 'ospray-studio'
		self.port = 8000

		self.config = config
		self.data_sources = data_sources

		self.service_command = (
			'docker service create '
			'--name ospray_studio '
			'--publish 80:5000/tcp '
			f'--replicas {self.config.get("aws", {}).get("replicas", 1)} '
			'--mount type=bind,src=/mnt/efs/data,dst=/data '
			'seelab/substrate-ospray-studio:latest '
			'flask run --host=0.0.0.0'
		)

	def start(self):
		mounts = super().start()

		docker = from_env()

		self.port = self.config['docker'].get('port', self.port)
		docker.services.create(
			'seelab/substrate-ospray-studio:latest',
			'flask',
			args=['run', '--host=0.0.0.0'],
			endpoint_spec=EndpointSpec(ports={self.port: (5000, 'tcp')}),
			mode=ServiceMode(
				mode='replicated',
				replicas=self.config['docker'].get('replicas', 1)
			),
			mounts=mounts,
			name='ospray-studio',
			networks=[f'substrate-{self.name}-net']
		)



import os
from pathlib import Path
import subprocess
import sys
import time

from docker import from_env
from docker.errors import APIError
import paramiko


class DockerSwarm():
	def __init__(self, _, config, tool):
		self.config = config
		self.tool = tool

		self.docker = from_env()
		self.network = None

		ssh_dir = os.path.join(Path.home(), '.ssh')
		ssh_dirfiles = os.listdir(ssh_dir)
		ssh_keys = [keyfile for keyfile in ssh_dirfiles if keyfile.startswith('id_')]
		ssh_pkeys = [keyfile for keyfile in ssh_keys if not keyfile.endswith('.pub')]
		self.key_paths = [os.path.join(ssh_dir, keyfile) for keyfile in ssh_pkeys]

	def create_swarm(self):
		self.log('Initializing the swarm…')
		advertise_addr = self.config['docker'].get('advertise_addr', None)
		try:
			if advertise_addr:
				self.docker.swarm.init(advertise_addr=advertise_addr)
			else:
				self.docker.swarm.init()
		except APIError:
			pass
		self.log('✓\n')

		self.network = self.docker.networks.create(
			f'substrate-{self.tool.name}-net',
			driver='overlay'
		)

		manager_token = self.docker.swarm.attrs['JoinTokens']['Manager']
		worker_token = self.docker.swarm.attrs['JoinTokens']['Worker']

		nodes = []
		for node in self.config['docker'].get('managers', []):
			nodes.append(('manager', node))
		for node in self.config['docker'].get('workers', []):
			nodes.append(('worker', node))

		for node_type, node in nodes:
			username, location = node.split('@')
			if node_type == 'manager':
				token = manager_token
			elif node_type == 'worker':
				token = worker_token

			self.log(f'Adding remote {node} to swarm as {node_type}…')

			with paramiko.client.SSHClient() as ssh_client:
				ssh_client.load_system_host_keys()
				ssh_client.connect(
					location,
					username=username,
					key_filename=self.key_paths
				)

				_, _, stderr = ssh_client.exec_command(
					f'docker swarm join --token {token} {advertise_addr}'
				)
				stderr = stderr.read()
				if stderr:
					self.log(f'✕\nremote error ({location}): {stderr}\n')
				else:
					self.log('✓\n')

		self.log('Creating the swarm service…')
		self.tool.start()
		self.log('✓\n')

		self.log('Waiting for the service to be stable…')
		output = subprocess.check_output(
			f'docker service ps {self.tool.name} --format ' + '"{{.CurrentState}}"',
			shell=True
		).decode('utf-8').split('\n')
		invalid_outputs = [line for line in output if len(line) > 0 and not line.startswith('Running')]  # noqa: E501
		while invalid_outputs:
			time.sleep(1)
			output = subprocess.check_output(
				f'docker service ps {self.tool.name} --format ' + '"{{.CurrentState}}"',
				shell=True
			).decode('utf-8').split('\n')
			invalid_outputs = [line for line in output if len(line) > 0 and not line.startswith('Running')]  # noqa: E501
		self.log('✓\n')

		return f'127.0.0.1:{self.tool.port}'

	def destroy_swarm(self):
		nodes = []
		for node in self.config['docker'].get('managers', []):
			nodes.append(node)
		for node in self.config['docker'].get('workers', []):
			nodes.append(node)

		for node in nodes:
			username, location = node.split('@')

			self.log(f'Removing remote {node} from swarm…')

			with paramiko.client.SSHClient() as ssh_client:
				ssh_client = paramiko.client.SSHClient()
				ssh_client.load_system_host_keys()
				ssh_client.connect(
					location,
					username=username,
					key_filename=self.key_paths
				)

				_, _, stderr = ssh_client.exec_command('docker swarm leave --force')
				stderr = stderr.read()
				if stderr:
					self.log(f'✕\nremote error ({location}): {stderr}\n')
				else:
					self.log('✓\n')

		self.log('Removing the swarm service…')
		for obj in self.docker.services.list(filters={'name': self.tool.name}) + self.docker.networks.list(names=[f'substrate-{self.tool.name}-net']):  # noqa: E501
			obj.remove()
		self.log('✓\n')

	def log(self, message):
		print(message, end='')
		sys.stdout.flush()

	def start(self):
		return self.create_swarm()

	def stop(self):
		self.destroy_swarm()



import os
import subprocess
from time import sleep

from aws_cdk import App, Environment, RemovalPolicy, Stack
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_efs as efs
from aws_cdk import aws_elasticloadbalancingv2 as elb
from aws_cdk import aws_iam as iam
from aws_cdk.aws_elasticloadbalancingv2_targets import InstanceTarget
import requests


class AWSStack():
	def __init__(self, path, config, tool):
		self.path = path
		self.config = config
		self.tool = tool
		self.region = self.config['aws'].get('region', 'us-east-1')  # noqa: E501
		self.use_https = self.config['aws'].get('https', False)

		app = App()
		_AWSStack(
			app,
			f'substrate-stack-{self.tool.name}',
			tool,
			config,
			self.tool.data_sources[1],
			env=Environment(region=self.region)
		)
		app.synth()

	def start(self):
		subprocess.run(
			f'npx cdk bootstrap --app "substrate {self.tool.name} -c {self.path} synth"',  # noqa: E501
			check=True,
			shell=True
		)

		subprocess.run(
			f'npx cdk deploy --require-approval never --app "substrate {self.tool.name} -c {self.path} synth"',  # noqa: E501
			check=True,
			shell=True
		)

		# TODO: this needs to be different for HTTPS
		location = subprocess.check_output(
			f'aws ec2 describe-instances --region {self.region} --filters Name=instance-state-name,Values=running Name=tag:Name,Values=substrate-stack-{self.tool.name}/substrate-leader --query Reservations[*].Instances[*].[PublicIpAddress] --output text',  # noqa: E501
			shell=True
		).strip().decode('utf-8')

		print('The CloudFormation stack has successfully deployed.')  # noqa: E501
		print('\033[91mIt may take several minutes before the instance is ready to use, please wait while the instance starts.\033[0m')  # noqa: E501
		while True:
			print('Checking if AWS instance is ready…', end='')
			try:
				response = requests.get(f'http://{location}')
				if response.status_code == 200:
					break
			except requests.exceptions.ConnectionError:
				pass

			print("instance isn't ready yet. Trying again in 30 seconds.")
			sleep(30)
		print('✓')

		return location

	def stop(self):
		subprocess.run(
			f'npx cdk destroy --force --app "substrate {self.tool.name} -c {self.path} synth"',  # noqa: E501
			check=True,
			shell=True
		)
		subprocess.run(
			f'aws s3 rm s3://{self.config["aws"]["bucket"]} --recursive',
			check=True,
			shell=True
		)


class _AWSStack(Stack):  # pylint: disable=too-many-instance-attributes
	def __init__(self, scope, _id, tool, config, data_urls, **kwargs):
		self.tool = tool
		self.config = config
		self.data_urls = data_urls
		self.leader_name = ''
		self.nodes = []
		self.use_https = self.config['aws'].get('https', False)

		self.tool.upload_to_s3()

		super().__init__(scope, _id, **kwargs)

		self.vpc = ec2.Vpc(
			self,
			_id,
			max_azs=2 if self.use_https else 1,
			nat_gateways=0,
			subnet_configuration=[
				ec2.SubnetConfiguration(name='public', subnet_type=ec2.SubnetType.PUBLIC)
			]
		)

		ami_string = self.config['aws'].get('ami', None)
		if ami_string:
			self.ami = ec2.MachineImage.generic_linux({
				self.region: 'ami-048ff3da02834afdc'
			})
		else:
			self.ami = ec2.MachineImage.latest_amazon_linux(
				generation=ec2.AmazonLinuxGeneration.AMAZON_LINUX_2
			)

		role_arn = self.config['aws'].get('role_arn', None)
		if role_arn:
			self.role = iam.Role.from_role_arn(self, 'john', role_arn, mutable=False)
		else:
			self.role = None

		self.file_system = self.provision_fs()
		self.provision_ec2()
		if self.use_https:
			self.provision_elb()

	def add_leader_commands(self, udata, *args):
		for command in args:
			udata.add_commands(command)

			if self.config['aws'].get('save_logs', True):
				udata.add_commands(f'aws s3 sync /var/log s3://{self.config["aws"]["bucket"]}/logs')  # noqa: E501

	def get_udata(self, _type):
		udata = ec2.UserData.for_linux()
		self.add_leader_commands(
			udata,
			f'export AWS_DEFAULT_REGION={self.region}',  # noqa: E501
			f'export AWS_ACCESS_KEY_ID={os.environ.get("AWS_ACCESS_KEY_ID")}',
			f'export AWS_SECRET_ACCESS_KEY={os.environ.get("AWS_SECRET_ACCESS_KEY")}',
			f'export AWS_SESSION_TOKEN={os.environ.get("AWS_SESSION_TOKEN", "")}',
			'sudo yum check-update -y',
			'sudo yum upgrade -y',
			'sudo amazon-linux-extras install docker',
			'sudo service docker start',
			'sudo usermod -a -G docker ec2-user',
			'sudo yum install -y amazon-efs-utils',
			'sudo yum install -y nfs-utils',
			'sudo yum install -y python3',
			'sudo mkdir -p "/mnt/efs"',
			f'test -f "/sbin/mount.efs" && echo "{self.file_system.file_system_id}:/ /mnt/efs efs defaults,_netdev" >> /etc/fstab || echo "{self.file_system.file_system_id}.efs.{self.region}.amazonaws.com:/ /mnt/efs nfs4 nfsvers=4.1,rsize=1048576,wsize=1048576,hard,timeo=600,retrans=2,noresvport,_netdev 0 0" >> /etc/fstab',  # noqa: E501
			'mount -a -t efs,nfs4 defaults',
			'until mountpoint -d /mnt/efs; do mount -a -t efs,nfs4 defaults; done'
		)

		if _type == 'leader':
			self.add_leader_commands(
				udata,
				'sudo mkdir -p /etc/pki/tls/private',
				'cd /etc/pki/tls/private',
				'sudo openssl req -x509 -newkey rsa:4096 -nodes -out cert.pem -keyout key.pem -days 365 -subj "/C=US/ST=Tennessee/L=Knoxville/O=University of Tennessee/OU=Seelab/CN=github.com\/seelabutk"'  # pylint: disable=anomalous-backslash-in-string # noqa: E501,W605
			)

			self.add_leader_commands(
				udata,
				f'aws s3 sync s3://{self.config["aws"]["bucket"]} /mnt/efs',
				'cd /mnt/efs/data',
				*[f'curl -O {data_url}' for data_url in self.data_urls],
				'cd -'
			)

			self.add_leader_commands(
				udata,
				'mkdir /mnt/efs/swarm',
				'docker swarm init',
				'docker swarm join-token worker | grep "docker" > /mnt/efs/swarm/worker',
				'docker swarm join-token manager | grep "docker" > /mnt/efs/swarm/manager',
				self.tool.service_command
			)
		elif _type == 'manager':
			udata.add_commands('until [ -f /mnt/efs/swarm/manager ]; do sleep 1; done')
			udata.add_commands('$(cat /mnt/efs/swarm/manager)')
		elif _type == 'worker':
			udata.add_commands('until [ -f /mnt/efs/swarm/worker ]; do sleep 1; done')
			udata.add_commands('$(cat /mnt/efs/swarm/worker)')

		return udata

	def provision_ec2(self):
		managers = self.config['aws'].get('managers', {})
		workers = self.config['aws'].get('workers', {})

		for index, _type in enumerate(managers):
			count = managers.get(_type, 1)
			for index2 in range(count):
				if index == 0 and index2 == 0:
					instance_name = 'substrate-leader'
					udata = self.get_udata('leader')
				else:
					instance_name = f'substrate-manager-{_type}-{index}'
					udata = self.get_udata('manager')

				self.nodes.append(ec2.Instance(
					self,
					instance_name,
					instance_type=ec2.InstanceType(instance_type_identifier=_type),
					machine_image=self.ami,
					role=self.role,
					user_data=udata,
					vpc=self.vpc,
					vpc_subnets=ec2.SubnetSelection(subnet_type=ec2.SubnetType.PUBLIC)
				))

		for _type in workers:
			udata = self.get_udata('worker')

			count = workers.get(_type, 1)
			for index in range(count):
				self.nodes.append(ec2.Instance(
					self,
					f'substrate-worker-{_type}-{index}',
					instance_type=ec2.InstanceType(instance_type_identifier=_type),
					machine_image=self.ami,
					role=self.role,
					user_data=udata,
					vpc=self.vpc,
					vpc_subnets=ec2.SubnetSelection(subnet_type=ec2.SubnetType.PUBLIC)
				))

		for node in self.nodes:
			node.node.add_dependency(self.file_system.mount_targets_available)

			if self.role:
				node.instance.iam_instance_profile = None
				node.node.try_remove_child('InstanceProfile')

			# TODO: fix
			node.connections.allow_from_any_ipv4(
				ec2.Port.all_traffic(),
				'General Purpose'
			)

	def provision_elb(self):
		load_balancer = elb.ApplicationLoadBalancer(
			self,
			'substrate-elb',
			vpc=self.vpc
		)
		listener = load_balancer.add_listener('substrate-elb-listener', port=80)
		listener.add_targets(
			'substrate-ec2-targets',
			port=80,
			targets=[InstanceTarget(instance) for instance in self.nodes]
		)

	def provision_fs(self):
		efs_sg = ec2.SecurityGroup(self, 'substrate-efs-sg', vpc=self.vpc)
		efs_sg.add_ingress_rule(
			peer=ec2.Peer.any_ipv4(),  # TODO: fix
			connection=ec2.Port.tcp(2049)
		)

		return efs.FileSystem(
			self,
			'substrate-efs',
			removal_policy=RemovalPolicy.DESTROY,
			security_group=efs_sg,
			vpc=self.vpc
		)



import os
import subprocess

from docker import from_env
from docker.types import Mount
from docker.types.services import EndpointSpec, ServiceMode

from . import Tool


class Tapestry(Tool):  # pylint: disable=too-many-instance-attributes
	def __init__(self, config, data_sources):
		super().__init__(config, data_sources)

		self.name = 'tapestry'
		self.config = config
		self.port = 8000

		self.service_command = (
			'docker service create '
			'--env APP_DIR=/app '
			'--name tapestry '
			'--publish 80:9010/tcp '
			f'--replicas {self.config.get("aws", {}).get("replicas", 1)} '
			'--mount type=bind,src=/mnt/efs/app,dst=/app '
			'--mount type=bind,src=/mnt/efs/config,dst=/config '
			'--mount type=bind,src=/mnt/efs/data,dst=/data '
			'seelab/substrate-tapestry:latest '
			'./server /config 9010 /app'
		)

		fallback_dir = os.path.join(os.path.dirname(__file__), 'tapestry')

		self.tapestry_path = self.config['tapestry'].get('directory', [fallback_dir])[0]  # noqa: E501
		self.app_path = os.path.join(self.tapestry_path, 'app')
		self.config_path = os.path.join(self.tapestry_path, 'config')
		self.data_sources = data_sources

	def start(self):
		mounts = super().start()
		docker = from_env()

		mounts.append(Mount('/app', self.app_path, type='bind', read_only=True))
		mounts.append(
			Mount('/config', self.config_path, type='bind', read_only=True)
		)

		self.port = self.config['docker'].get('port', self.port)
		docker.services.create(
			'seelab/substrate-tapestry:latest',
			'./server',
			args=['/config', '9010', '/app'],
			endpoint_spec=EndpointSpec(ports={self.port: (9010, 'tcp')}),
			env=['APP_DIR=/app'],
			mode=ServiceMode(
				mode='replicated',
				replicas=self.config['docker'].get('replicas', 1)
			),
			mounts=mounts,
			name='tapestry',
			networks=[f'substrate-{self.name}-net']
		)

	def upload_to_s3(self):
		super().upload_to_s3()

		subprocess.run(
			f'aws s3 sync {self.app_path} s3://{self.config["aws"]["bucket"]}/app',
			check=True,
			shell=True
		)

		subprocess.run(
			f'aws s3 sync {self.config_path} s3://{self.config["aws"]["bucket"]}/config',  # noqa: E501
			check=True,
			shell=True
		)



import subprocess

from docker.types import Mount


class Tool():
	def __init__(self, config, data_sources):  # pylint: disable=unused-argument
		# Provide a name for this tool
		self.name = ''

		self.config = config
		self.data_sources = data_sources

		# Define the Docker CLI command to create this service on AWS.
		self.service_command = ''

	# Start a local Docker swarm service
	def start(self):
		mounts = []
		data_paths = self.data_sources[0]
		if len(data_paths) > 1:
			for index, data_path in enumerate(data_paths):
				mounts.append(
					Mount(f'/data/{index}', data_path, type='bind', read_only=True)
				)
		else:
			mounts.append(Mount('/data', data_paths[0], type='bind', read_only=True))

		return mounts

	# Upload files needed to run your service to S3.
	# Upload them to mirror the directory structure you want on EFS.
	def upload_to_s3(self):
		print('Syncing data to S3')

		region = self.config['aws'].get('region', 'us-east-1')
		output = subprocess.check_output('aws s3 ls', shell=True).decode('utf-8')

		if self.config['aws']['bucket'] not in output:
			subprocess.run(
				f'aws s3 mb s3://{self.config["aws"]["bucket"]} --region { region}',
				check=True,
				shell=True
			)

		for data_path in self.data_sources[0]:
			subprocess.run(
				f'aws s3 sync {data_path} s3://{self.config["aws"]["bucket"]}/data',
				check=True,
				shell=True
			)



import os
import subprocess

from docker import from_env
from docker.types import Mount
from docker.types.services import EndpointSpec, ServiceMode

from . import Tool


class Braid(Tool):
	def __init__(self, config, data_sources):
		super().__init__(config, data_sources)

		self.name = 'braid'
		self.config = config
		self.port = 8000

		self.braid_pattern = self.config['braid']['file_pattern']

		self.service_command = (
			'docker service create '
			f'--env VCI_PATTERN={self.braid_pattern} '
			'--env VCI_ROOT=/data '
			'--name tapestry '
			'--publish 80:8840/tcp '
			f'--replicas {self.config.get("aws", {}).get("replicas", 1)} '
			'--mount type=bind,src=/mnt/efs/data,dst=/data '
			'--mount type=bind,src=/mnt/efs/app,dst=/opt/run '
			'-w /opt/run'
			'seelab/substrate-braid:latest '
			'python3.7 -u -m vci'
		)

		fallback_dir = os.path.join(os.path.dirname(__file__), 'braid')

		self.braid_path = self.config['braid'].get('directory', fallback_dir)
		self.data_sources = data_sources

	def start(self):
		mounts = super().start()

		docker = from_env()

		mounts.append(
			Mount('/opt/run', self.braid_path, type='bind', read_only=True)
		)

		self.port = self.config['docker'].get('port', self.port)
		docker.services.create(
			'seelab/substrate-braid:latest',
			'python3.7',
			args=['-u', '-m', 'vci'],
			endpoint_spec=EndpointSpec(ports={self.port: (8840, 'tcp')}),
			env=[
				f'VCI_PATTERN={self.braid_pattern}',
				'VCI_ROOT=/data'
			],
			mode=ServiceMode(
				mode='replicated',
				replicas=self.config['docker'].get('replicas', 1)
			),
			mounts=mounts,
			name='braid',
			networks=[f'substrate-{self.name}-net'],
			workdir='/opt/run'
		)

	def upload_to_s3(self):
		super().upload_to_s3()

		subprocess.run(
			f'aws s3 sync {self.braid_path} s3://{self.config["aws"]["bucket"]}/app',
			check=True,
			shell=True
		)



from docker import from_env
from docker.types.services import EndpointSpec, ServiceMode

from . import Tool


class NetCDFSlicer(Tool):
	def __init__(self, config, data_sources):
		super().__init__(config, data_sources)

		self.name = 'nc_slicer'
		self.port = 8000

		self.config = config
		self.data_sources = data_sources

		self.service_command = (
			'docker service create '
			'--name dchm '
			'--publish 80:5000/tcp '
			f'--replicas {self.config.get("aws", {}).get("replicas", 1)} '
			'--mount type=bind,src=/mnt/efs/data,dst=/data '
			'seelab/substrate-nc-slicer'
			'python app.py'
		)

	def start(self):
		mounts = super().start()
		docker = from_env()

		docker.services.create(
			'seelab/substrate-nc-slicer',
			endpoint_spec=EndpointSpec(ports={self.port: (5000, 'tcp')}),
			mode=ServiceMode(
				mode='replicated',
				replicas=self.config['docker'].get('replicas', 1)
			),
			mounts=mounts,
			name='nc-slicer',
			networks=[f'substrate-{self.name}-net'],
			init=True
		)

	# TODO: 
	# fix nc-slicer -> nc_slicer and init=True in main substrate repo
	# set up dockerhub org and move images to it
	# impliment water-and-land changes to nc_slicer image 




#!/usr/bin/env python3
from argparse import ArgumentParser
import os
import sys
from urllib.request import urlretrieve
from urllib.parse import urlparse

import yaml

from .targets import AWSStack, DockerSwarm
from .tools import HelloWorld, NetCDFSlicer, OSPRayStudio, Tapestry, Braid

TOOLS = {
	'hello-world': HelloWorld,
	'nc_slicer': NetCDFSlicer,
	'ospray-studio': OSPRayStudio,
	'tapestry': Tapestry,
	'braid': Braid
}


class Substrate():
	def __init__(self, tool_name, path=None):
		self.tool_name = tool_name
		self.path, self.config = self._parse_yaml(path)

		self.target = None
		if self.config.get('aws', None) is not None:
			self.target = AWSStack
		elif self.config.get('docker', None) is not None:
			self.target = DockerSwarm

		self.data_sources = self._get_data(self.config)
		if tool_name not in TOOLS:
			raise Exception(f'No tool named {tool_name}')
		self.tool = TOOLS[tool_name](self.config, self.data_sources)

		self.target_obj = self.target(self.path, self.config, self.tool)

	def _get_data(self, config):
		source_paths = config['data']['source']
		data_paths = []
		data_urls = []

		for source_path in source_paths:
			is_url = urlparse(source_path).scheme.startswith(('ftp', 'http'))

			# Download the dataset if necessary to the target location
			if is_url and self.target == DockerSwarm:
				target_path = os.path.abspath(config['data']['target'])

				os.makedirs(os.path.dirname(target_path), exist_ok=True)
				urlretrieve(source_path, target_path)

				if target_path not in data_paths:
					data_paths.append(target_path)
			elif is_url:
				data_urls.append(source_path)
			else:
				data_paths.append(os.path.abspath(source_path))

			if data_paths and os.path.isfile(data_paths[-1]):
				data_paths[-1] = os.path.dirname(data_paths[-1])

		return (data_paths, data_urls)

	def _parse_yaml(self, path):
		config_name = 'substrate.config.yaml'

		if path is None:
			path = os.getcwd()

			files = os.listdir(path)
			if config_name not in files:
				raise FileNotFoundError(config_name)

			path = os.path.join(path, config_name)

		with open(path, 'r', encoding='utf8') as stream:
			_config = yaml.load(stream, Loader=yaml.Loader)

		if _config.get('docker', None) and _config.get('aws', None):
			sys.exit('The "docker" and "aws" options cannot be used simultaneously.')

		return (path, _config)

	def start(self):
		return f'http://{self.target_obj.start()}'

	def stop(self):
		self.target_obj.stop()


def main():
	parser = ArgumentParser(description='Launches a Substrate instance')

	parser.add_argument(
		'tool',
		help='The visualization tool to run [hello_world, nc_slicer, ospray_studio, tapestry, braid]',  # noqa: E501
		metavar='TOOL'
	)
	parser.add_argument(
		'action',
		help='[start, stop]',
		metavar='ACTION'
	)
	parser.add_argument('-c', '--config', dest='path')
	args = parser.parse_args()

	if args.action not in ['start', 'stop', 'synth']:
		sys.exit('Invalid action specified.')

	substrate = Substrate(args.tool, args.path)
	if args.action == 'start':
		print(f'You may view your new visualization stack here: {substrate.start()}.')  # noqa: E501

	if args.action == 'stop':
		substrate.stop()
